#!/usr/dist/pkgs/perl,v5.003/5bin.sun4/perl
require 5.003;

BEGIN {
    my($workdir) = ($0 =~ /(.*)\/.*/);
    unshift(@INC,"$workdir/../lib");
}
my($workdir) = ($0 =~ /(.*)\/.*/);

# Load libraries.
use BER;
use SNMP_Session;
use SNMP_MIB;
use SNMP_Simple;
use Socket;
use strict;
use Ping;

# Given an IP address string and a netmask string, return the
# string with the network address for that net.
sub getnet($$) {
    my ($ipaddr, $netmask) = @_;
    my $ipnum = inet_aton($ipaddr);
    
    # inet_aton has trouble with the all ones netmask.  Just return
    # the raw IP address.
    if ($netmask eq "255.255.255.255") { return $ipaddr; }

    my $netmasknum = inet_aton($netmask);

    # Catch problems with translating netmask and IP address here ...
    if (!defined($netmasknum) || !defined($ipnum)) {
	print LOG "Something is undefined: netmask $netmask or IP $ipaddr\n";
	return $ipaddr;
    }

    my $netnum = $ipnum & $netmasknum;

    # And just in case there still are problems ...
    if (!defined($netnum)) {
	print LOG "Netnum not defined with addr $ipaddr mask $netmask\n";
	return $ipaddr;
    }

    return inet_ntoa($netnum);    
}

# Configure the loaded libraries ... don't need warnings.
$SNMP_Session::suppress_warnings = 1;
$SNMP_Simple::suppress_warnings = 1;

SNMP_MIB::loadoids("$workdir/../mibs", "RFC1213");

### MAIN ###
# Grab and parse arguments.
# Arguments: <hostname> <network> <netmask>
# <hostname> is the name of a router to start the discover with.
# <network>/<netmask> is the network to find the topology of.

my $host = $ARGV[0];
my $network = $ARGV[1];
my $netmask = $ARGV[2];

# Check and see if the arguments are valid.
my $hostip = gethostbyname($host);
die "Invalid hostname: $host" unless defined $hostip;
my $startip = inet_ntoa($hostip);

my $netnum = inet_aton($network);
die "Invalid network: $network" unless defined $netnum;
$network = inet_ntoa($netnum);

my $netmasknum = inet_aton($netmask);
die "Invalid netmask: $netmask" unless defined $netmasknum;
$netmask = inet_ntoa($netmasknum);

# Misc. initalization
my $ping = Ping->new("icmp");
my $community = "public";
my $basedir = "/var/tmp";
$| = 1;

# Initalize logs
open(LOG, ">$basedir/discover.$network.log");
autoflush LOG 1;

print LOG "Discover started at ".time."\n";
print LOG "Starting: $startip $network $netmask\n";

open (RTR, ">$basedir/discover.$network.routers");
autoflush RTR 1;

print "graph SWAN { \n\tnode [ fontsize=12, shape=box ]\n";

my (%routerlist, %routervisited, @worklist, $curhost);
$routerlist{$startip} = $host;

# Main loop
for ($curhost = $startip; defined($curhost); $curhost = shift @worklist) {

    my $hostname = gethostbyaddr(pack('C4', split(/\./, $curhost)), AF_INET);
    
    next if (defined($routervisited{$curhost}));
    
    print LOG "Visiting $curhost\n";
    $routervisited{$curhost} = 1;

    # First, check if you can ping this host.  
    if ($ping->ping($curhost)) {
	
	# Ok, ping worked, now try SNMP
	my $session = SNMP_Simple->open($curhost, $community, 161);

        if (!defined($session)) {
	    print LOG "Failed open of SNMP session to host $curhost\n";
	    next;
	}

	$session->{'session'}->{'retries'} = 5;

	# Get basic system info
	my ($sysName, $sysContact, $sysLocation, $sysObjectID, $router) 
	    = $session->snmpget('sysName', 'sysContact', 'sysLocation', 
				'sysObjectID', 'ipForwarding');

	if (!defined($sysName) || ($sysName eq "")) {
	    print LOG "$curhost does not have sysName defined\n";
	    $sysName = $curhost;
	}

	# Check if it is SNMP enabled and is a router.
	if (defined($router) && ($router == 1)) {
	    print RTR "$sysName -> $curhost\n";
	    
	    # Get IP address table entries.
	    my @rtrips = $session->snmpgettable('ipAdEntAddr');
	    my %ipindicies = $session->snmpgettablea('ipAdEntIfIndex');
	    my %netmasks = $session->snmpgettablea('ipAdEntNetMask');
	    my $indexoid = SNMP_MIB::name2oid('ipAdEntIfIndex');
	    my $netoid = SNMP_MIB::name2oid('ipAdEntNetMask');
	    my %intsvisited;
	    my $rtrip;

	    print "\t\"$sysName\" -- {";

	    # Iterate through the IP addresses, and make a graph
	    # connection for each.  Mark the interface as visited, 
	    # so we do not make another connection for that interface
	    # later when we are looking at the routes.

	    foreach $rtrip (@rtrips) {
		my $lnetmask = $netmasks{"$netoid.$rtrip"};
		my $index = $ipindicies{"$indexoid.$rtrip"};
		if (!defined($rtrip)) {
		    print LOG "IP not defined?  $curhost is confused.\n";
		    next;
		}
		if (!defined($lnetmask)) {
		    print LOG "Netmask not defined for $curhost IP $rtrip interface $index\n";
		    next;
		}
		


		# check if this is the loopback net (127.0.0.0/8)
		# and go on if it is.
		my $loopback = getnet($rtrip, "255.0.0.0");
		next if ($loopback eq "127.0.0.0");

		$routerlist{$rtrip} = $sysName;

		# If we have not already visited this IP, mark it as
		# visited.
		if (!defined($routervisited{$rtrip})) {
		    $routervisited{$rtrip} = 1;
		}

		my $net = getnet($rtrip, $lnetmask);

		# Skip printing out this network if it is not in
		# the overall net we are interested in.
		my $checknet = getnet($rtrip, $netmask);
		next if ($checknet ne $network);

		# Mark this as visited so that the loop below won't
		# draw a connection between this router and a neighbor
		# router.

		$intsvisited{$index} = 1;

		print "\"$net\" ; ";
	    }

	    # Get route table information.
	    my @routedests = $session->snmpgettable('ipRouteDest');
	    my %routeindicies = $session->snmpgettablea('ipRouteIfIndex');
	    my %nexthops = $session->snmpgettablea('ipRouteNextHop');
	    my $routeoid = SNMP_MIB::name2oid('ipRouteIfIndex');
	    my $hopoid = SNMP_MIB::name2oid('ipRouteNextHop');
	    my ($dest, %routers, %uniqrouters);

	    # Iterate through the route table.
	    foreach $dest (@routedests) {
		my $routeindex = $routeindicies{"$routeoid.$dest"};
		my $nextrouter = $nexthops{"$hopoid.$dest"};
		my $loopbacknet = getnet($dest, "225.0.0.0");
		my $checknet = getnet($dest, $netmask);
		
		# Check for good SNMP returns.
		if (!defined($nextrouter) || ($nextrouter eq "")) {
		    print LOG "next hop from $curhost to $dest is not defined\n";
		    next;
		}
		
		# Ignore the route if it is to the loopback net, if it
		# is a default route, if it goes back to this host, or
		# if the route goes somewhere not in the network we want
		# to check.  Also, ignore summary routes where the route
		# is to the whole network, and not a subnet.
		next if (($checknet ne $network) || 
			 ($dest eq $network) ||
			 ($loopbacknet eq "127.0.0.0") || 
			 ($nextrouter eq "0.0.0.0") ||
			 ($nextrouter eq $curhost));

		# Check if this list is on the global list of routers,
		# and add it if it is not.
		if (!defined($routerlist{$nextrouter})) {
		    $routerlist{$nextrouter} = $nextrouter;
		    push @worklist, $nextrouter;
		    print LOG "Added $nextrouter to the worklist\n";
		}

		# If we have already seen this interface before when
		# looking at IP addresses, don't bother to do anything
		# more.  Also, if we have already seen this router,
		# don't do anything more.
		if (!defined($intsvisited{$routeindex})
		    && !defined($routers{$nextrouter})) {

		    my $nextsession = SNMP_Simple->open($nextrouter, 
							$community, 161);

		    # Bail and go to the next destination if
		    # we can not open a session.
		    next if (!defined($nextsession));

		    my ($nextsysName) = $nextsession->snmpget('sysName');
		    
		    # Check for a valid return.
		    if (defined($nextsysName) && ($nextsysName ne "")) {
			$routers{$nextrouter} = $nextsysName;

			# Check to see if the system name has
			# been seen before, if not print it out.
			if (!defined($uniqrouters{$nextsysName})) {
			    if ($nextsysName ne $sysName) {
				print "\"$nextsysName\" ; ";
			    }
			    $uniqrouters{$nextsysName} = 1;
			}

		    } else {
			$nextsession->close();
			next;
		    }
		    $nextsession->close();
		}

	    } # end foreach $dest
	    print "}\n";

	} else { # router not defined
	    print LOG "$curhost does not respond to SNMP or is not a router\n";
	} # end if (!defined($router) ...

	$session->close();
    } else { # can not ping host
	print LOG "$curhost not reachable\n";
    } # end if ($ping->ping($curhost))

} # end for $curhost

$ping->close();
close(LOG);
close(RTR);

print "}\n";

exit 0;
